@@ -1,9 +1,14 @@
 import inspect
 
+
 def get_default_args(func):
     signature = inspect.signature(func)
     return {
         k: v.default
         for k, v in signature.parameters.items()
         if v.default is not inspect.Parameter.empty
-    }
\ No newline at end of file
+    }
+
+
+if __name__=="__main__":
+    add_default_models()
\ No newline at end of file
